%
%   PCC2 Standard Library
%
%   This file defines commands and keys used internally by PCC2,
%   and a number of user-callable routines described in the
%   manual. Do not modify unless you are absolutely 100% positively
%   sure what you are doing.
%
%   This file will be overwritten by PCC2 upgrades. It will most
%   likely only work with the PCC2 version it was intended for.
%
%   This version of core.q is for PCC2, and is incompatible with
%   PCC 1.x - and vice versa.
%

%Option Optimize(2), Lexical

% @q CC$LibraryVersion:Str (Internal)
% Version of the standard library (<tt>core.q</tt>).
Dim Shared CC$LibraryVersion = '1.99.26'

%%% Console-Mode Replacements for GUI routines %%%%%%%%%%%%%%%%%%%%%%

If Not System.GUI Then

  %%% Utilities

  % Return the index of the first word in list which starts with the given word.
  Function CC$MatchWord (list, word)
    Local i:=1, e
    Do While list<>''
      e := First(' ', list)
      If Mid(e,1,Len(word)) = word Then Return i
      i := i + 1
      list := Trim(Rest(' ', list))
    Loop
  EndFunction

  % Count number of words in list
  Function CC$CountWords (list)
    Local i:=0
    Do While list<>''
      i:=i+1
      list:=Trim(Rest(' ', list))
    Loop
    Return i
  EndFunction

  % Concatenate all strings in an array
  Function CC$ConcatArray(a)
    Local i, result=''
    For i:=0 To Dim(a)-1 Do
      result := result # a(i)
    Next
    Return result
  EndFunction


  %%% Actual User Functions

  % MessageBox generates a console message.
  Sub MessageBox (text, Optional heading)
    % FIXME: PCC 1.x replaces \r by \n in text.
    If IsEmpty(heading) Then heading := Translate("Message")
    Print heading, ": ", text
  EndSub

  % UI.Update, UI.PopupConsole has no effect in console mode.
  Sub UI.Update (Optional flag)
  EndSub

  Sub UI.PopupConsole
  EndSub

  % Commands that call a dialog generate an error in console mode.
  Sub UI.ChooseObject (screen)
    Abort "Not in graphics mode"
  EndSub

  Sub UI.Search (query, Optional flags)
    Abort "Not in graphics mode"
  EndSub

  Sub UI.ListShips (X, Y, Optional flags, ok, Heading)
    Abort "Not in graphics mode"
  EndSub

  Sub UI.ListFleets (X, Y, Optional flags, ok, Heading)
    Abort "Not in graphics mode"
  EndSub

  Sub UI.GotoScreen (sid, Optional oid)
    Abort "Not in graphics mode"
  EndSub

  Sub UI.GotoChart (X, Y)
    Abort "Not in graphics mode"
  EndSub

  Sub UI.SelectionManager
    Abort "Not in graphics mode"
  EndSub

  Sub UI.PlanetInfo (pid)
    Abort "Not in graphics mode"
  EndSub

  Sub UI.KeymapInfo
    Abort "Not in graphics mode"
  EndSub

  Sub UI.Help (page)
    Abort "Not in graphics mode"
  EndSub

  % This differs from PCC1 behaviour, which will allow you to create the list box,
  % but fail when you try to invoke it.
  Function Listbox (title, Optional current, width, height, help)
    Abort "Not in graphics mode"
  EndFunction

  % UI.InputNumber generates an error in console mode (like it does in PCC 1.x,
  % although we can probably do better).
  Sub UI.InputNumber (prompt, Optional lo, hi, cur, help)
    Abort "Not in graphics mode"
  EndSub

  Sub UI.InputFCode (flags, Optional def)
    UI.Input Translate("Change Friendly Code"), "", 3, "h", def
  EndSub

  Sub UI.FileWindow (title, wildcard, Optional Help)
    UI.Input title
  EndSub

  % UI.Message generates a console message and a prompt
  Sub UI.Message (text, Optional heading, buttons)
    MessageBox text, Heading
    If IsEmpty(buttons) Then buttons := Translate("OK")
    Do
      UI.Input buttons
      If IsEmpty(UI.Result) Then
        % This means the last choice
        UI.Result := Z(CC$CountWords(buttons))
        Break
      Else If UI.Result = "" Then
        % This means the first choice
        UI.Result := 1
        Break
      Else
        % Does this match any button?
        UI.Result := CC$MatchWord(buttons, UI.Result)
        If Not IsEmpty(UI.Result) Then Break
      EndIf
    Loop
  EndSub

  % UI.OverlayMessage generates a console message
  Sub UI.OverlayMessage (text)
    Print Translate("Message") # ": " # text
  EndSub

  % Rich Text operations
  % Note that these are not a perfect emulation: IsString(RAdd('x'))=True with these
  % functions, but IsString(RAdd('x'))=False with the GUI builtins
  Function RAdd(args())
    Return CC$ConcatArray(args)
  EndFunction

  Function RMid(text, i, n)
    Return Mid(text, i, n)
  EndFunction

  Function RString(text)
    Return text
  EndFunction

  Function RLen(text)
    Return Len(text)
  EndFunction

  Function RStyle(style, args())
    Return CC$ConcatArray(args)
  EndFunction

  Function RLink(link, args())
    Return CC$ConcatArray(args)
  EndFunction

  Function RXml(xml, args())
    Local res
    Local state = 0        % 0=idle, 1=opening tag, 2=opening tag+"", 3=opening tag +'', 4=&, 5=failed &
    Local argidx, i, ch
    Local i
    If Not IsEmpty(xml) Then
      % Parse it
      res := ''
      For i:=1 To Len(xml) Do
        ch := Mid(xml,i,1)
        Select Case state
          Case 0
            If ch = '&'  Then
              state := 4
              argidx := 0
            Else If ch = '<' Then
              state := 1
            Else
              res := res & ch
            EndIf
          Case 1
            If ch = "\"" Then state := 2
            If ch = "'"  Then state := 3
            If ch = '>'  Then state := 0
          Case 2
            If ch = "\"" Then state := 0
          Case 3
            If ch = "'"  Then state := 0
          Case 4
            If ch >= '0' And ch <= '9' Then
              argidx := 10*argidx + Asc(ch) - Asc('0')
            Else If ch = ';' Then
              If argidx >= 0 And argidx < Dim(args) Then res := res & args(argidx)
              state := 0
            Else
              state := 5
            EndIf
          Case 5
            % FIXME: it would probably make sense to parse well-known entities as well,
            % which are &lt;, &#x1234; etc.
            If ch = ';' Then state := 0
        EndSelect
      Next
    EndIf
    Return res
  EndFunction

  % LoadResource as no effect.
  Function LoadResource(name)
  EndFunction
EndIf


%%% User-callable Subroutines %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% @q WaitOneTurn (Global Command)
% Suspend script for one turn.
% Execution proceeds next turn (or later, when a turn is missed).
% See {Stop} for details and restrictions on suspension.
% @see Stop
% @since PCC2 1.99.10, PCC 1.0.6
Sub WaitOneTurn ()
  Local t = Turn
  Do While t = Turn
    Stop
  Loop
EndSub

% @q MoveTo x:Int, y:Int (Ship Command)
% Move ship to X,Y.
% This sets a waypoint, and suspends the script until the waypoint has been reached.
% As a special case, if the waypoint is in a warp well, the script continues when
% the planet has been reached.
%
% Example:
% <pre class="ccscript">
%   MoveTo 1240, 1500
%   MoveTo 1400, 1520
% </pre>
% This will move the ship to (1240,1500) and then to (1400,1520).
%
% This command is intended to be used in auto tasks.
% @see Stop
% @since PCC2 1.99.10, PCC 1.0.6
Sub MoveTo (x, y)
  % We terminate when the waypoint is zero: when moving across a map border, we never
  % reach the specified x,y.
  Local pid = PlanetAt(x, y, 1)
  With Lock("s" & Id & ".waypoint")
    SetWaypoint x, y
    Do Until (Waypoint.DX=0 And Waypoint.DY=0) Or ((pid<>0) And (Orbit$=pid))
      Stop
    Loop
  EndWith
EndSub

% @q MoveTowards x:Int, y:Int (Ship Command)
% Move ship towards X,Y.
% This command sets the waypoint and waits one turn.
% This causes the ship to move towards the specified coordinate, but not necessarily reach it.
%
% This command is intended to be used in auto tasks.
% @see Stop
% @since PCC2 1.99.10, PCC 1.1.3
Sub MoveTowards (x, y)
  With Lock("s" & Id & ".waypoint")
    SetWaypoint x, y
    WaitOneTurn
  EndWith
EndSub

% @q EnqueueShip h:Int, e:Int, Optional bt:Int, bc:Int, tt:Int, tc:Int (Planet Command)
% Build a ship. Parameters are the same as for %BuildShip.
% Whereas %BuildShip immediately replaces the build order,
% this command waits until the starbase is done building its current order.
% This command also waits until sufficient resources are available to place the order.
%
% This command is intended to be used in auto tasks.
% @see BuildShip, Stop
% @since PCC2 1.99.10, PCC 1.0.6
Sub EnqueueShip (h, e, Optional bt, bc, tt, tc)
  % wait until pending build order performed
  Do While Build.YesNo
    Stop
  Loop
  % wait until build order successfully delivered
  Do
    Try
      BuildShip h, e, bt, bc, tt, tc
      Break
    EndTry
    Stop
  Loop
EndSub

% @q CargoUploadWait amount:Cargo, Optional flags:Str (Ship Command)
% Load cargo from planet, wait until amount loaded.
% This command is similar to {CargoUpload}, see there for details about its parameters.
% It will try to load cargo from the planet this ship is orbiting.
% It will continue execution when enough cargo has been loaded.
%
% For example,
% | CargoUpload "200n"
% will load 200 kt Neutronium on this ship.
% If the planet has more than 200 kt Neutronium, they will be loaded immediately;
% if the planet has less, the ship will wait for newly-mined Neutronium, or Neutronium unloaded by ships,
% until 200 kt have been loaded.
% If this ship has not enough free cargo room, it will also wait until some becomes available.
%
% Note that when you unload cargo from this ship while %CargoUploadWait is active,
% that cargo will be uploaded again and count towards the limit.
% If the planet has 100 kt fuel while you try to load 200, everything will be uploaded.
% If you now unload these 100 kt again, %CargoUploadWait will load them up again and finish successfully,
% as it now has loaded 200 kt total.
% @see CargoUpload, CargoTransferWait
% @since PCC 1.0.11, PCC2 1.99.21
Sub CargoUploadWait (cargo, Optional flags)
  Local Cargo.Remainder = cargo
  Do
    CargoUpload Cargo.Remainder, flags & "n"
    If Not Cargo.Remainder Then Return
    Stop
  Loop
EndSub

% @q CargoTransferWait amount:Cargo, target:Int, Optional flags:Any (Ship Command, Planet Command)
% Transfers cargo to a ship, wait until amount transferred.
% This command is similar to {CargoTransfer}, see there for details about its parameters.
% It will try to transfer cargo from this unit to ship %sid.
% It will continue execution when all cargo has been transferred.
%
% For example,
% | CargoTransfer "200n", 42
% will load 200 kt Neutronium onto ship 42.
% If the unit this command is invoked from has enough cargo, and the receiving ship has enough free room,
% the command will succeed immediately.
% Otherwise, the script will wait until cargo or free room becomes available,
% and continue until all cargo has been transferred.
% @see CargoUpload, CargoUploadWait
% @since PCC 1.0.11, PCC2 1.99.21
Sub CargoTransferWait (sid, cargo, Optional flags)
  Local Cargo.Remainder = cargo
  Do
    CargoTransfer sid, Cargo.Remainder, flags & "n"
    If Not Cargo.Remainder Then Return
    Stop
  Loop
EndSub


% @q CargoUnloadAllShips (Ship Command, Planet Command)
% Unload all ships at this location. Can be called from ship or planet.
% @since PCC2 1.99.16
Sub CargoUnloadAllShips
  Local thisX = Loc.X, thisY = Loc.Y
  Local Cargo.Remainder
  If Not PlanetAt(thisX, thisY) Then
    Abort "No planet at this location"
  Else
    ForEach Ship Do
      If Loc.X=thisX And Loc.Y=thisY And Played Then
        CargoUnload "10000tdmcs$", "n"
      EndIf
    Next
  EndIf
EndSub


% @q BuildMinesWait amount:Int (Planet Command)
% Build mines, wait as necessary.
% If %amount mines cannot be built immediately due to lacking resources or colonists,
% this command waits until they have been built.
% @see BuildMines, Stop
% @since PCC2 1.99.10, PCC 1.0.17
Sub BuildMinesWait (amount)
  Local Build.Remainder = amount
  Do
    BuildMines Build.Remainder, "n"
    If Not Build.Remainder Then Return
    Stop
  Loop
EndSub

% @q BuildFactoriesWait amount:Int (Planet Command)
% Build factories, wait as necessary.
% If %amount factories cannot be built immediately due to lacking resources or colonists,
% this command waits until they have been built.
% @see BuildFactories, Stop
% @since PCC2 1.99.10, PCC 1.0.17
Sub BuildFactoriesWait (amount)
  Local Build.Remainder = amount
  Do
    BuildFactories Build.Remainder, "n"
    If Not Build.Remainder Then Return
    Stop
  Loop
EndSub

% @q BuildDefenseWait amount:Int (Planet Command)
% Build defense posts, wait as necessary.
% If %amount defense posts cannot be built immediately due to lacking resources or colonists,
% this command waits until they have been built.
% @see BuildDefense, Stop
% @since PCC2 1.99.10, PCC 1.0.17
Sub BuildDefenseWait (amount)
  Local Build.Remainder = amount
  Do
    BuildDefense Build.Remainder, "n"
    If Not Build.Remainder Then Return
    Stop
  Loop
EndSub

% @q BuildBaseDefenseWait amount:Int (Planet Command)
% Build starbase defense, wait as necessary.
% If %amount defense posts cannot be built immediately due to lacking resources or colonists,
% this command waits until they have been built.
% @see BuildBaseDefense, Stop
% @since PCC2 1.99.10, PCC 1.0.17
Sub BuildBaseDefenseWait (amount)
  Local Build.Remainder = amount
  Do
    BuildBaseDefense Build.Remainder, "n"
    If Not Build.Remainder Then Return
    Stop
  Loop
EndSub

% @q SetFleetName name:Str (Ship Command)
% Change fleet name. This ship must be member of a fleet to use this command.
% The fleet's name will be changed to %name.
% @since PCC2 1.99.17, PCC 1.0.14
Sub SetFleetName (name)
  If Not IsEmpty(name) Then
    If Fleet$=0 Then Abort "Not in a fleet"
    Ship(Fleet$).Fleet.Name := name
  EndIf
EndSub

% @q ChangeFleetLeader fid:Int, sid:Int (Global Command)
% Change fleet leader. %fid is the fleet Id (and thus the Id of the fleet's leader).
% %sid is the Id of the ship that will become the new leader of the fleet.
% That ship should already be a member of the fleet.
% @since PCC2 1.99.17, PCC 1.0.14
Sub ChangeFleetLeader (fid, sid)
  If fid=sid Then Return
  % start the new fleet
  With Ship(sid) Do
    SetFleet sid
    Fleet.Name := Ship(fid).Fleet.Name
  EndWith
  % First move all members to the new fleet, then move the old leader
  ForEach Ship Do If (Id<>fid) And (Fleet$=fid) Then SetFleet sid
  With Ship(fid) Do SetFleet sid
EndSub

% @q SelectionLayer l:Any (Global Command)
% Change active selection layer.
% This is a convenience routine that accepts the parameter in multiple formats, namely
% a selection layer number (0-7) or name ("A"-"H").
% @since PCC2 1.99.10, PCC 1.0.15
Sub SelectionLayer (l)
  If IsEmpty(l) Then Return
  If IsString(l) Then l:=InStr("abcdefgh", l)-1
  Selection.Layer := l
EndSub

% @q FixShipWait (Ship Command)
% @q FixShipWait id:Int (Planet Command)
% Fix ship at starbase when possible.
% Waits until the order can be submitted (i.e. the starbase's shipyard is idle), then gives the order.
%
% Although documented as a Ship and Planet command, this actually is a global command.
% When calling it from a ship, do not specify a parameter.
% When calling it from a planet, specify a ship Id.
% @since PCC2 1.99.21, PCC 1.0.16
% @see FixShip (Ship Command), FixShip (Planet Command)
Sub FixShipWait (Optional id)
  % The ship detection differs between PCC 1.x and PCC2.
  If Type.Short<>"p" Then
    % assume ship context
    If (Not Orbit$) Or (Not Planet(Orbit$).Base.YesNo) Then Abort "Not at a starbase"
    Do While Planet(Orbit$).Shipyard
      Stop
    Loop
    FixShip
  Else
    % assume starbase context
    Do While Shipyard
      Stop
    Loop
    FixShip id
  EndIf
EndSub

% @q RecycleShipWait (Ship Command)
% @q RecycleShipWait id:Int (Planet Command)
% Recycle ship at starbase when possible.
% Waits until the order can be submitted (i.e. the starbase's shipyard is idle), then gives the order.
%
% Although documented as a Ship and Planet command, this actually is a global command.
% When calling it from a ship, do not specify a parameter.
% When calling it from a planet, specify a ship Id.
% @since PCC2 1.99.21, PCC 1.0.16
% @see RecycleShip (Ship Command), RecycleShip (Planet Command)
Sub RecycleShipWait (Optional id)
  If Type.Short<>"p" Then
    % assume ship context
    If (Not Orbit$) Or (Not Planets(Orbit$).Base.YesNo) Then Abort "Not at a starbase"
    Do While Planets(Orbit$).Shipyard
      Stop
    Loop
    RecycleShip
  Else
    % assume starbase context
    Do While Shipyard
      Stop
    Loop
    RecycleShip id
  EndIf
EndSub

% @q Notify s:Str, Optional t:Int (Global Command)
% Notify message.
% This command stops execution of an auto task, and informs the player of an event using the message text %s.
% These messages are displayed in the notification message window,
% from which users can resume execution of an auto task, or stop it.
% The text will be word-wrapped to fit in the window;
% it can also contain carriage returns to force line breaks.
%
% The message window will automatically include the name of the object the
% auto task belongs to, so the message need not repeat that. For example,
% <pre class="ccscript">
%   Notify "out of fuel"
% </pre>
% will generate the following message:
% <pre>
%   (-s0124)<<< Notify >>>
%   FROM: Auto Task Ship #124
%
%   out of fuel
% </pre>
%
% When the turn number %t is specified, the message will pop up in the specified turn
% (and execution of the calling script will suspend that long),
% otherwise it will be sent immediately.
%
% A message sent using %Notify will remain active until it is confirmed.
% If the player does not confirm a message during this PCC session, it will pop up again in the next session.
% @see AddNotify
% @since PCC2 1.99.16, PCC 1.0.18
Sub Notify (s, Optional t)
  % This is implemented totally different from PCC 1.x.
  % This command initially appeared in 1.0.16, where it was morally equivalent to 'Print'.
  If Not IsEmpty(t) Then
    Do While Turn < t
      Stop
    Loop
  EndIf
  Do
    CC$Notify s, True
    Stop
  Loop While Not CC$NotifyConfirmed()
EndSub

% @q AddNotify s:Str (Global Command)
% Notify message.
% This command informs the player of an event using the message text %s.
% These messages are displayed in the notification message window.
% The text will be word-wrapped to fit in the window;
% it can also contain carriage returns to force line breaks.
%
% The message window will automatically include the name of the object the
% auto task belongs to, so the message need not repeat that. For example,
% <pre class="ccscript">
%   AddNotify "out of fuel"
% </pre>
% will generate the following message:
% <pre>
%   (-s0124)<<< Notify >>>
%   FROM: Auto Task Ship #124
%
%   out of fuel
% </pre>
%
% Messages sent using %AddNotify will only be visible for the PCC session in which the command was used.
% Unlike %Notify, execution of the script will automatically continue after the command,
% the message needs not be confirmed by the user.
% @see Notify
% @since PCC2 1.99.16, PCC 1.1.16
Sub AddNotify (s)
  CC$Notify s, False
EndSub

% @q CC$AutoReCheck:void (Internal)
% Part of the implementation of {Restart}. A call of this routine is generated before the loop jump.
% This routine makes sure we do not enter infinite loops.
Sub CC$AutoReCheck
  % Create process-local variable upon first entry into this routine
  Dim Static CC$AutoReState
  % Error when we're getting here again with a turn number we've already seen
  If CC$AutoReState = Turn
    Notify Translate("This Auto Task performed a \"Restart\" loop without intervening delays. You should probably use \"WaitOneTurn\".")
    CC$AutoReState := 0
  Else
    CC$AutoReState := Turn
  EndIf
EndSub

% @q CC$AutoExec cmd:Str (Internal)
% Execute auto task command. An auto task command <tt>foo</tt> is
% actually coded as <tt>CC$AutoExec "foo"</tt>. This implements error handling.
Sub CC$AutoExec(cmd)
  Do
    Try
      Eval cmd
      Return
    Else
      Notify Translate("Error during execution of auto task: ") + System.Err
    EndTry
  Loop
EndSub

% Sub AutoLoad (file, name, Optional args)

% @q FindShipCloningAt(pid:Int):Int (Function)
% Find ship cloning at a planet.
% Returns the Id of the ship that is cloning at planet %pid.
% If no ship is trying to clone, or cloning is forbidden, returns EMPTY.
% @since PCC2 1.99.10
Function FindShipCloningAt (pid)
  If Cfg("AllowShipCloning")
    % FindShip returns the first ship, which is just what we need
    Return FindShip(Orbit$=pid And StrCase(FCode="cln"))
  Else
    Return Z(0)
  EndIf
EndFunction

% @q ShipNameAndId(sid:Int):Str (Function)
% Get name of a ship, together with its Id.
%
% Note that this function does what %ShipName does in PCC 1.x.
% In PCC2, %ShipName returns just the plain name which I believe is more useful
% and consistent with %PlanetName.
% Code that used %ShipName can now use %ShipNameAndId.
% @since PCC2 1.99.20
% @see ShipName
Function ShipNameAndId (sid)
  Local N = ShipName(sid)
  If N
    Return N & " (#" & sid & ")"
  Else
    Return Format(Translate("Ship %d"), sid)
  EndIf
EndFunction

%%% GUI Subroutines %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

If System.GUI Then
%% When the GUI is not active, don't define all these to save memory

Dim Shared CC$BuildH, CC$BuildE, CC$BuildBT, CC$BuildBC, CC$BuildTT, CC$BuildTC

% Ctrl-P on starbase/ship build screen
Sub CC$PickupBuild
  If Build.YesNo Then
    CC$BuildH  := Build.Hull$
    CC$BuildE  := Build.Engine$
    CC$BuildBT := Build.Beam$
    CC$BuildBC := Build.Beam.Count
    CC$BuildTT := Build.Torp$
    CC$BuildTC := Build.Torp.Count
    Messagebox Format(Translate("Ship specification for a %s saved. Use [Ctrl-B] on other starbases to build this ship there."), Build.Short)
  EndIf
EndSub

% Ctrl-B on starbase screen
Sub CC$PerformBuild
  Local System.Err
  If IsEmpty(CC$BuildH) Or Not Played Then Return
  If Build.YesNo Then
    UI.Message Format(Translate("This starbase is already building a %s. Override this with the saved build order for a %s?"), Build.Short, Hull(CC$BuildH).Name.Short), Translate("Ship Building"), Translate("Yes No")
    If UI.Result<>1 Then Return
  EndIf
  Try
    BuildShip CC$BuildH, CC$BuildE, CC$BuildBT, CC$BuildBC, CC$BuildTT, CC$BuildTC
  Else
    Messagebox Format(Translate("Unable to build a %s: %s"), Hull(CC$BuildH).Name.Short, System.Err)
  EndTry
EndSub

% Ctrl-S on ship build screen (not yet documented because it's weird).
% This is still the inelegant PCC 1.x version; we could do better by
% using a user-defined function.
Sub CC$SearchBuild
  % Value of tech level t = 50*(t-1)*t
  If Build.YesNo Then
    Local mstr := Hull(Build.Hull$).Cost.Str
    Local tstr := "Max(0,Tech.Hull*(Tech.Hull-1)-" & (Hull(Build.Hull$).Tech.Hull*(Hull(Build.Hull$).Tech.Hull-1)) & ")"
    mstr := CAdd(mstr, CMul(Engine(Build.Engine$).Cost.Str, Hull(Build.Hull$).Engine.Count))
    tstr := tstr & "+Max(0,Tech.Engine*(Tech.Engine-1)-" & (Engine(Build.Engine$).Tech*(Engine(Build.Engine$).Tech-1)) & ")"
    If Build.Beam.Count Then
      mstr := CAdd(mstr, CMul(Beam(Build.Beam$).Cost.Str, Build.Beam.Count))
      tstr := tstr & "+Max(0,Tech.Beam*(Tech.Beam-1)-" & (Beam(Build.Beam$).Tech*(Beam(Build.Beam$).Tech-1)) & ")"
    EndIf
    If Build.Torp.Count Then
      mstr := CAdd(mstr, CMul(Torpedo(Build.Torp$).Cost.Str, Build.Torp.Count))
      tstr := tstr & "+Max(0,Tech.Torpedo*(Tech.Torpedo-1)-" & (Torpedo(Build.Torp$).Tech*(Torpedo(Build.Torp$).Tech-1)) & ")"
    EndIf
    UI.Search "Not Build.YesNo And CCompare(Mined.Str&(Supplies+Money)&'$','" & mstr & "'&(" & tstr & ")*50&'$')", "2b"
  EndIf
EndSub

% '.' everywhere
Sub CC$ToggleSelection
  Mark Not Marked
EndSub

% F9 on ship
Sub CC$SetShipComment
  UI.Input Translate("Enter new comment for this ship:"), Translate("Edit Comment"), 255, "gm30", Comment
  SetComment UI.Result
EndSub

% F9 on planet
Sub CC$SetPlanetComment
  UI.Input Translate("Enter new comment for this planet:"), Translate("Edit Comment"), 255, "gm30", Comment
  SetComment UI.Result
EndSub

% Shift-F9 on control screens
Sub CC$SetPlanetCommentAt(X,Y)
  Local pid = PlanetAt(X,Y,1)
  If pid Then With Planet(pid) Do CC$SetPlanetComment
EndSub

% N in ship screen
Sub CC$ShipName
  If Not Played Then Return
  UI.Input Translate("Enter new name:"), Format(Translate("Rename Starship #%d"), Id), 20, "gm20", Name
  SetName UI.Result
EndSub

% Sub CC$FleetName
% Sub CC$ChangeLeader

% C-z
Sub CC$SetWaypoint (x, y)
  Try With Lock("s" & Id & ".waypoint") Do SetWaypoint x, y
EndSub

% Alt-Left/Alt-Right
Sub CC$SwitchSelection (delta)
  Selection.Layer := (Selection.Layer + 8 + delta) Mod 8
  UI.OverlayMessage Format(Translate("Selection %c"), Chr(Selection.Layer + 65))
EndSub

% Shift/Ctrl-F2/F3/F8
Sub CC$PlanetAt (typ, X, Y)
  Local pid := PlanetAt(X, Y, 1)
  If pid Then Try UI.GotoScreen typ, pid
EndSub

% Shift/Ctrl-F4
Sub CC$GotoChart (X, Y)
  If IsEmpty(X) Or IsEmpty(Y) Then
    UI.GotoScreen 4
  Else
    UI.GotoChart X,Y
  EndIf
EndSub

% L, Shift-L, Shift-F1, etc.
Sub CC$ListShips (X, Y, flags)
  Local UI.Result
  UI.ListShips X, Y, flags
  If UI.Result
    UI.GotoScreen 1, UI.Result
  EndIf
EndSub

% C-n
Sub CC$ListShipPrediction (X, Y, Optional Id)
  Local UI.Result
  UI.ListShipPrediction X, Y, Id
  If UI.Result
    UI.GotoScreen 1, UI.Result
  EndIf
EndSub

% X on ship screen
Sub CC$ShipXchg
  Local myX = UI.X, myY = UI.Y, myId = Id
  If FindShip(Loc.X=myX And Loc.Y=myY And Id<>myId And Played)
    % There is one of our ships other than this one at this location
    CC$ListShips myX, myY, 'e' & myId
  Else
    % No ship, so check for a planet
    Local pid = PlanetAt(myX, myY, 1)
    If pid And Planet(pid).Played Then UI.GotoScreen 2, pid
  EndIf
EndSub

% X on planet screen
Sub CC$PlanetXchg
  Local pid = PlanetAt(UI.X, UI.Y, 1)
  If pid And pid<>Id And Planet(pid).Played Then
    UI.GotoScreen 2, pid
  Else
    If FindShip(Loc.X=UI.X And Loc.Y=UI.Y And Played) Then CC$ListShips UI.X, UI.Y, 'e'
  EndIf
EndSub

% X on base screen
Sub CC$BaseXchg
  Local pid = PlanetAt(UI.X, UI.Y, 1)
  If pid And pid<>Id And Planet(pid).Played Then
    If Planet(pid).Base.YesNo Then
      UI.GotoScreen 3, pid
    Else
      UI.GotoScreen 2, pid
    EndIf
  Else
    If FindShip(Loc.X=UI.X And Loc.Y=UI.Y And Played) Then CC$ListShips UI.X, UI.Y, 'e'
  EndIf
EndSub

% F1/F2/F3
Sub CC$GotoScreen (Screen)
  Local UI.Result
  UI.ChooseObject Screen
  If Not IsEmpty(UI.Result) Then UI.GotoScreen Screen, UI.Result
EndSub

% F1/F2/F3 on starchart
Sub CC$GotoObject (Screen, Id)
  If Id Then Try UI.GotoScreen Screen, Id
EndSub

% F1 on history screen
Sub CC$ShipScreenFromHistory
  If Played Then
    UI.GotoScreen 1, Id
  Else
    CC$GotoScreen 1
  EndIf
EndSub

% F5
Sub CC$PlanetInfo (X, Y)
  Local pid = PlanetAt(X,Y,True)
  If pid Then UI.PlanetInfo pid
EndSub

% F8
Sub CC$BuildOrGotoBase
  If Base.YesNo
    Try UI.GotoScreen 3, Id
  Else
    Try CC$BuildBase
  EndIf
EndSub

% ESC on race screen
Sub CC$ExitRace
  Local UI.Result
  UI.Message Translate("Do you want to exit this game?"), Translate("Exit Game"), Translate("Yes No")
  If UI.Result=1 Then System.ExitRace
EndSub

% ESC on race screen
Sub CC$ExitClient
  Local UI.Result
  UI.Message Translate("Do you want to exit PCC2?"), Translate("PCC2"), Translate("Yes No")
  If UI.Result=1 Then System.ExitClient
EndSub

% Scanner movement
Sub CC$ScannerMoveX (delta)
  If UI.Prefix Then delta := If(delta < 0, -UI.Prefix, UI.Prefix)
  UI.X := UI.X + delta
EndSub

Sub CC$ScannerMoveY (delta)
  If UI.Prefix Then delta := If(delta < 0, -UI.Prefix, UI.Prefix)
  UI.Y := UI.Y + delta
EndSub

% B on starchart in planet view
Sub CC$BaseLock
  If Base.YesNo Then
    Chart.SetView 'BaseLock'
  Else
    Try CC$BuildBase
  EndIf
EndSub

% P on starchart in starbase view
Sub CC$PlanetLock
  Chart.SetView 'PlanetLock'
EndSub

% ESC on starchart. Since this hides the view, it will also turn off the keymap,
% so the next ESC will exit the map.
Sub CC$HidePanel
  Chart.SetView ''
EndSub

% C-p/C-s on ship screen. Only review the transfer if it is present.
Sub CC$ReviewShipTransfer (kind, Id)
  If Id Then CC$TransferShip kind, Id
EndSub

% c on ship/history screen
Sub CC$ShipCargo
  If Played Then
    CC$TransferShip 0,0
  Else
    CC$CargoHistory
  EndIf
EndSub

% u on ship screen
Sub CC$ShipUnload
  If Played Then
    If Orbit$ Then
      CC$TransferUnload
    Else
      MessageBox Translate("We are not orbiting a planet, sir. If you want to jettison cargo, use the 'c' command instead."), Translate("Cargo Transfer")
    EndIf
  EndIf
EndSub

% Ship operation
% FIXME: when the waypoint is tied to the scanner, and keyboard move is enabled,
% this triggers on every movement.
Sub CC$WithShipWaypoint (cmd, opt, Optional overrideAT)
  Local UI.Result
  Local sid = Id
  Local s, T
  If IsEmpty(overrideAT) Then overrideAT:=False

  % Check waypoint lock
  s := "s" & sid & ".waypoint"
  If Not overrideAT And GetLockInfo(s, 0) Then
    If GetLockInfo(s, 2) Then
      UI.Message Format(Translate("%s (%s)\nProceed anyway?"), GetLockInfo(s, 2), GetLockInfo(s, 1)), Translate("Waypoint"), Translate("Yes No")
    Else
      UI.Message Format(Translate("This function is under control of \"%s\".\nProceed anyway?"), GetLockInfo(s, 1)), Translate("Waypoint"), Translate("Yes No")
    EndIf
    If UI.Result <> 1 Then Return
  EndIf

  % Check fleet
  If Fleet$<>0 And Fleet$<>sid Then
    UI.Message Translate("This ship is member of a fleet. Do you want to set the waypoint of the whole fleet, or do you want to leave the fleet?"), Translate("Waypoint"), Translate("Set Leave Cancel")
    If UI.Result=3 Then Return
    If UI.Result=1 Then
      % Set
      sid:=Fleet$
    Else
      % Leave fleet
      Fleet$:=0
    EndIf
  EndIf

  % All tests passed
  With Ship(sid) Do Eval cmd

  % Set optimum speed
  If opt Then
    If Ship(sid).Speed<>"Hyperdrive" And Ship(sid).Waypoint.Dist>0 Then
      % Find out how long it takes to get there at this speed
      s := Global.Engine(Ship(sid).Engine$).Speed$
      Ship(sid).Speed$ := s
      T := Ship(sid).Move.Eta
      If T < 30 Then % FIXME: hardcoded value
        % Find whether we can reach the target with a slower speed
        Do While s > 1
          s := s-1
          Ship(sid).Speed$ := s
          If Ship(sid).Move.Eta > T Then
            % It's slower, so undo and stop
            Ship(sid).Speed$ := s+1
            Break
          EndIf
        Loop
      EndIf
    EndIf
  EndIf
EndSub

Sub CC$WithFleetWaypoint (cmd, opt)
  With Ship(Fleet$) Do CC$WithShipWaypoint cmd, opt
EndSub

% 'j' on fleet screen
Sub CC$FleetJoin
  Local UI.Result
  Local fid = Fleet$
  UI.ListFleets Ship(fid).Loc.X, Ship(fid).Loc.Y, "a" & fid, Translate("OK"), Translate("Join Fleet")
  If Not IsEmpty(UI.Result) Then
    % Move members, then leader
    ForEach Ship Do If Id<>fid And Fleet$=fid Then Fleet$:=UI.Result
    Ship(fid).Fleet$:=UI.Result
  EndIf
EndSub

% 'n' on fleet screen
Sub CC$FleetName
  Local UI.Result
  Local fid = Fleet$
  UI.Input Translate("Enter new name:"), Format(Translate("Rename Fleet #%d"), fid), 255, "gm20", Ship(fid).Fleet.Name
  SetFleetName UI.Result
EndSub

% 'p' on fleet screen
Sub CC$FleetSplit
  Local UI.Result
  Local oldfid = Fleet$                       % old fleet Id
  Local newmark = Not Ship(oldfid).Marked     % selection status for new fleet
  Local newcount = 0                          % number of ships in new fleet
  Local newfid = If (Marked=newmark, Id, 0)   % new leader is this ship if it has the right status

  % Count ships
  ForEach Ship Do
    If Fleet$=oldfid And Marked=newmark Then
      newcount:=newcount+1
      If newfid=0 Then newfid:=Id
    EndIf
  Next
  If newcount=0 Then
    MessageBox Translate("Please select the ships which should go into one fleet, and unselect those which should go into the other."), Translate("Split Fleet")
    Return
  EndIf

  % Confirm
  UI.Message Format(Translate("Form a new fleet with %d ship%!1{s%}, leader is %s?"), newcount, Ship(newfid).Name), Translate("Split Fleet"), Translate("Yes No")
  If UI.Result<>1 Then Return

  % Do it
  Ship(newfid).Fleet$ := newfid
  ForEach Ship Do If Fleet$=oldfid And Marked=newmark Then Fleet$:=newfid

  % Bring user there
  If Fleet$=newfid Then Iterator(10).CurrentIndex:=newfid
EndSub

% C-ins on fleet screen
Sub CC$FleetAddToSim
  Local fid = Fleet$
  ForEach Ship Do
    If Fleet$=fid Then CC$AddToSim False
  Next
EndSub

% DEL on fleet screen
Sub CC$FleetDeleteMember
  Local UI.Result
  If Fleet$ = Id Then
    % This is the fleet leader, so we'd better ask
    UI.Message Translate("Remove leader from fleet?"), Translate("Fleet"), Translate("Yes No")
    If UI.Result<>1 Then Return
  EndIf
  Fleet$ := 0
EndSub

% C-DEL on fleet screen
Sub CC$FleetDeleteFleet
  Local UI.Result
  Local fid = Fleet$
  UI.Message Translate("Dissolve this fleet?"), Translate("Fleet"), Translate("Yes No")
  If UI.Result=1 Then
    ForEach Ship Do If Id<>fid And Fleet$=fid Then Fleet$:=0
    Ship(fid).Fleet$:=0
  EndIf
EndSub

% 'b' in fleet screen
Sub CC$ChangeLeader
  Local UI.Result
  If Fleet$ = Id Then Return      % ship is already the leader
  UI.Message Translate("Make this ship the leader of the fleet?"), Translate("Fleet"), Translate("Yes No")
  If UI.Result=1 Then ChangeFleetLeader Fleet$, Id
EndSub

% S-F10 on ship screen
Sub CC$ShipGoToFleet
  Local UI.Result
  If Fleet$ Then
    UI.GotoScreen 10, Fleet$
  Else
    If Not FindShip(Fleet$) Then
      UI.Message Translate("You have not yet set up a fleet to join. Use the \"Start New Fleet\" command on the [F10] menu to start one."), Translate("Join Fleet"), Translate("OK")
    Else
      UI.ListFleets Loc.X, Loc.Y, "a", Translate("OK"), Translate("Join Fleet")
      If UI.Result Then Fleet$ := UI.Result
    EndIf
  EndIf
EndSub

% C-m no ship screen: go to chunnel mate
Sub CC$GotoMate
  Local mate
  If HasFunction('Chunneling') Or HasFunction('ChunnelSelf') Or HasFunction('ChunnelOthers') Then
    mate := Val(FCode)
    If mate And Ship(mate).Played And (Ship(mate).HasFunction('Chunneling') Or Ship(mate).HasFunction('ChunnelTarget')) Then
      UI.GotoScreen 1, mate
    EndIf
  EndIf
EndSub

% Alt-m on ship screen: search ships that have us as chunnel mate
Sub CC$SearchMate
  If HasFunction("Chunneling") Or HasFunction("ChunnelTarget") Then
    UI.Search "(HasFunction('Chunneling') Or HasFunction('ChunnelSelf') Or HasFunction('ChunnelOthers')) And Val(FCode)=" & Id, "s2"
  EndIf
EndSub

% F on ship/planet/base screen
Sub CC$ChangeFCode
  Local UI.Result
  UI.InputFCode "d", FCode
  SetFCode UI.Result
EndSub

EndIf % System.GUI

%%% Initialisation %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Properties.
CreateShipProperty   Comment
CreatePlanetProperty Comment

% Keymaps. Note that those differ from PCC 1.x because we support multiple inheritance.
CreateKeymap Global, Ship, Planet, Base, Fleet

CreateKeymap ControlScreen(Global)
CreateKeymap ShipScreen(Ship, ControlScreen)
CreateKeymap PlanetScreen(Planet, ControlScreen)
CreateKeymap BaseScreen(Base, ControlScreen)
CreateKeymap FleetScreen(Fleet, ControlScreen)
CreateKeymap HistoryScreen(Ship, ControlScreen)
CreateKeymap AutoTaskScreen(ControlScreen)
CreateKeymap ShipTaskScreen(AutoTaskScreen)
CreateKeymap PlanetTaskScreen(AutoTaskScreen)
CreateKeymap BaseTaskScreen(AutoTaskScreen)

CreateKeymap Starchart(Global)
CreateKeymap ShipLock(Ship, Starchart)
CreateKeymap PlanetLock(Planet, Starchart)
CreateKeymap UnknownPlanetLock(PlanetLock)
CreateKeymap BaseLock(Base, Starchart)

CreateKeymap RaceScreen(Global)
CreateKeymap ShipBuildScreen(Global)

% Global Variables (System.Err is referenced by the interpreter core and therefore defined there).
% @q UI.Result:Str (Global Variable)
% Result of last user-interface operation.
% Operations like {UI.Message} store their result in a variable %UI.Result.
% If a local variable %UI.Result is visible, the result is stored in that instead of the global one.
% It is therefore a good idea to define a local variable to capture results without disturbing
% or being disturbed by other scripts.
% @assignable
Dim Shared UI.Result

% @q UI.Directory:Str (Global Variable)
% Current directory.
% This is the default directory for operations like {UI.FileWindow}.
% @assignable
Dim Shared UI.Directory

% @q UI.Prefix:Int (Global Variable)
% Current prefix argument.
% In keyboard actions (commands bound to a key using {Bind}),
% this variable contains the current <a href="prefixarg">prefix argument</a> (0 if none).
% @assignable
Dim Shared UI.Prefix

% @q Cargo.Remainder:Cargo (Global Variable)
% Remaining cargo after partial operation.
% If an operations like {CargoUnload} cannot finish completely, because of lacking cargo or cargo room,
% it can be told to do as much as possible instead of failing using the <tt>"n"</tt> flag.
% In this case, it stores the amount of unprocessed cargo in this variable.
%
% If a local variable %Cargo.Remainder is visible, the result is stored in that instead of the global one.
% It is therefore a good idea to define a local variable to capture results without disturbing
% or being disturbed by other scripts.
% @assignable
Dim Shared Cargo.Remainder

% @q Build.Remainder:Int (Global Variable)
% Remaining amount after partial build.
% If an operation like {BuildFactories} cannot finish completely, because of lacking resources or room,
% it can be told to build as much as possible instead of failing using the <tt>"n"</tt> flag.
% In this case, it stores the amount of structures not built in this variable.
%
% If a local variable %Build.Remainder is visible, the result is stored in that instead of the global one.
% It is therefore a good idea to define a local variable to capture results without disturbing
% or being disturbed by other scripts.
% @assignable
Dim Shared Build.Remainder

% Keybindings
Bind Base            "a"      := "CC$BuildAmmo"
Bind Base            "b"      := "CC$BuildShip"
Bind Base            "c-p"    := "CC$PickupBuild", "c-b" := "CC$PerformBuild"
Bind Base            "d"      := "CC$BuildStructures 2"
Bind Base            "f"      := "CC$ChangeFCode"
Bind Base            "m"      := "CC$ChangeMission"
Bind Base            "t"      := "CC$ChangeTech"
Bind Base            "c-c"    := "CC$GotoObject 1, FindShipCloningAt(Id)"
Bind Base            "ins"    := "Try CC$AddToSim"
Bind Base            "u"      := "CC$TransferPlanet 1"
Bind Base            "c"      := "CC$TransferPlanet 0"
Bind Base            "c-s"    := "CC$SearchBuild"             % FIXME: normally, this is on ShipBuildScreen only
Bind Base            "f9"     := "CC$SetPlanetComment"
Bind Base            "s-m"    := "CC$ViewMessages"
Bind BaseScreen      "x"      := "CC$BaseXchg"
Bind BaseScreen      "f8"     := "UI.GotoScreen 2, Id"
Bind BaseScreen      "c-r"    := "CC$GotoObject 1, Shipyard.Id"
Bind BaseScreen      "tab"    := "CC$BuildShip"
Bind BaseScreen      "s"      := "CC$ShipCostCalc"
Bind BaseScreen      "h"      := "UI.Help 'pcc2:basescreen'"
Bind BaseScreen      "alt-h"  := "UI.Help 'pcc2:basescreen'"
Bind BaseScreen      "ret"    := "CC$GotoObject 13, Id"
Bind BaseTaskScreen  "ret"    := "CC$GotoObject 3, Id"
Bind BaseTaskScreen  "f8"     := "UI.GotoScreen 2, Id"
Bind BaseTaskScreen  "f9"     := "CC$SetPlanetComment"
Bind BaseTaskScreen  "h"      := "UI.Help 'pcc2:basetaskscreen'"
Bind BaseTaskScreen  "alt-h"  := "UI.Help 'pcc2:basetaskscreen'"

Bind Planet          "c-s"    := "CC$BuySupplies"
Bind Planet          "f"      := "CC$ChangeFCode"
Bind Planet          "f9"     := "CC$SetPlanetComment"
Bind Planet          "g"      := "CC$Give"
Bind Planet          "s"      := "CC$SellSupplies"
Bind Planet          "t"      := "CC$ChangeTaxes"
Bind Planet          "ins"    := "Try CC$AddToSim"
Bind Planet          "f8"     := "CC$BuildOrGotoBase"
Bind Planet          "u"      := "Try CC$TransferPlanet 1"
Bind Planet          "c"      := "Try CC$TransferPlanet 0"
Bind Planet          "s-m"    := "CC$ViewMessages"
Bind PlanetScreen    "b"      := "CC$BuildStructures 0"
Bind PlanetScreen    "tab"    := "CC$BuildStructures 0"
Bind PlanetScreen    "d"      := "CC$BuildStructures 2"
Bind PlanetScreen    "m"      := "CC$BuildStructures 1"
Bind PlanetScreen    "x"      := "CC$PlanetXchg"
Bind PlanetScreen    "h"      := "UI.Help 'pcc2:planetscreen'"
Bind PlanetScreen    "alt-h"  := "UI.Help 'pcc2:planetscreen'"
Bind PlanetScreen    "ret"    := "CC$GotoObject 12, Id"
Bind PlanetTaskScreen "ret"   := "CC$GotoObject 2, Id"
Bind PlanetTaskScreen "f8"    := "CC$PlanetAt 3, Chart.X, Chart.Y"
Bind PlanetTaskScreen "f9"    := "CC$SetPlanetComment"
Bind PlanetTaskScreen "h"     := "UI.Help 'pcc2:planettaskscreen'"
Bind PlanetTaskScreen "alt-h" := "UI.Help 'pcc2:planettaskscreen'"

Bind Ship            "c-f"    := "Try SetFCode Planet(Orbit$).FCode"
Bind Ship            "c-z"    := "CC$SetWaypoint Loc.X, Loc.Y"
Bind Ship            "c"      := "CC$ShipCargo"
Bind Ship            "c-j"    := "Try CC$TransferShip 2, 0"
Bind Ship            "c-p"    := "CC$ReviewShipTransfer 2, Transfer.Unload.Id"
Bind Ship            "c-s"    := "CC$ReviewShipTransfer 1, Transfer.Ship.Id"
Bind Ship            "e"      := "CC$ChangePE"
Bind Ship            "f"      := "CC$ChangeFCode"
Bind Ship            "f9"     := "CC$SetShipComment"
Bind Ship            "g"      := "CC$Give"
Bind Ship            "m"      := "CC$ChangeMission"
Bind Ship            "s-m"    := "CC$ViewMessages"
Bind Ship            "n"      := "CC$ShipName"
Bind Ship            "r"      := "Try CC$RemoteControl"
Bind Ship            "s"      := "CC$ShipSpec"
Bind Ship            "u"      := "CC$ShipUnload"
Bind Ship            "c-i"    := "Try UI.GotoScreen 1, Mission.Intercept"
Bind Ship            "c-t"    := "Try UI.GotoScreen 1, Mission.Tow"
Bind Ship            "a-i"    := "UI.Search 'Mission.Intercept=' & Id, 's2'"
Bind Ship            "a-t"    := "UI.Search 'Mission.Tow=' & Id, 's2'"
Bind Ship            "ins"    := "Try CC$AddToSim"
Bind Ship            "c-z"    := "CC$WithShipWaypoint 'SetWaypoint Loc.X, Loc.Y', True"
Bind Ship            "s-f10"  := "CC$ShipGoToFleet"
Bind ShipScreen      "f10"    := 0                        % cancel global binding!
Bind ShipScreen      "s-f1"   := "CC$ListShips Chart.X, Chart.Y, 'e' & Id"
Bind ShipScreen      "c-f1"   := "CC$ListShips UI.X, UI.Y, 'e' & Id"
Bind ShipScreen      "c-n"    := "CC$ListShipPrediction UI.X, UI.Y, Id"
Bind ShipScreen      "x"      := "CC$ShipXchg"
Bind ShipScreen      "f8"     := "CC$PlanetAt 3, Chart.X, Chart.Y"
Bind ShipScreen      "h"      := "UI.Help 'pcc2:shipscreen'"
Bind ShipScreen      "alt-h"  := "UI.Help 'pcc2:shipscreen'"
Bind ShipScreen      "f6"     := "CC$GotoObject 6, Id"
Bind ShipScreen      "ret"    := "CC$GotoObject 11, Id"
Bind ShipScreen      "d"      := "CC$TransferMulti"
Bind ShipScreen      "w"      := "CC$WithShipWaypoint 'CC$ChangeSpeed', False"
Bind ShipScreen      "a"      := "CC$WithShipWaypoint 'CC$ChangeWaypoint', False"
Bind ShipScreen      "c-w"    := "CC$WithShipWaypoint 'SetWaypoint UI.X, UI.Y', True"
Bind ShipScreen      "dblclick" := "CC$WithShipWaypoint 'SetWaypoint UI.X, UI.Y', True"
Bind ShipScreen      "alt-m"  := "CC$SearchMate"
Bind ShipScreen      "c-m"    := "CC$GotoMate"
Bind ShipScreen      "q"      := "CC$ExplainPrediction"
Bind ShipTaskScreen  "ret"    := "CC$GotoObject 1, Id"
Bind ShipTaskScreen  "w"      := "CC$WithShipWaypoint 'CC$ChangeSpeed', False, True"
Bind ShipTaskScreen  "c"      := "CC$ShipCargo"
Bind ShipTaskScreen  "u"      := "CC$ShipUnload"
Bind ShipTaskScreen  "f6"     := "CC$GotoObject 6, Id"
Bind ShipTaskScreen  "f9"     := "CC$SetShipComment"
Bind ShipTaskScreen  "h"      := "UI.Help 'pcc2:shiptaskscreen'"
Bind ShipTaskScreen  "alt-h"  := "UI.Help 'pcc2:shiptaskscreen'"
Bind ShipTaskScreen  "f8"     := "CC$PlanetAt 3, Chart.X, Chart.Y"
Bind HistoryScreen   "f1"     := "CC$ShipScreenFromHistory"
Bind HistoryScreen   "f6"     := "CC$GotoScreen 6"
Bind HistoryScreen   "h"      := "UI.Help 'pcc2:historyscreen'"
Bind HistoryScreen   "alt-h"  := "UI.Help 'pcc2:historyscreen'"
Bind HistoryScreen   "x"      := "CC$ShipXchg"

Bind ControlScreen   "."      := "CC$ToggleSelection"
Bind ControlScreen   "pgup"   := "CC$CSBrowse 0,0", "-" := "CC$CSBrowse 0,0", "wheelup" := "CC$CSBrowse 0,0"
Bind ControlScreen   "pgdn"   := "CC$CSBrowse 1,0", "+" := "CC$CSBrowse 1,0", "wheeldn" := "CC$CSBrowse 1,0"
Bind ControlScreen   "c-pgup" := "CC$CSBrowse 0,1", "c--" := "CC$CSBrowse 0,1", "c-wheelup" := "CC$CSBrowse 0,1"
Bind ControlScreen   "c-pgdn" := "CC$CSBrowse 1,1", "c-+" := "CC$CSBrowse 1,1", "c-wheeldn" := "CC$CSBrowse 1,1"
Bind ControlScreen   "s-f9"   := "CC$SetPlanetCommentAt Loc.X, Loc.Y"
Bind ControlScreen   "c-f9"   := "CC$SetPlanetCommentAt UI.X, UI.Y"
Bind ControlScreen   "s-f1"   := "CC$ListShips Chart.X, Chart.Y, 'e'"
Bind ControlScreen   "c-f1"   := "CC$ListShips UI.X, UI.Y, 'e'"
Bind ControlScreen   "s-f2"   := "CC$PlanetAt 2, Chart.X, Chart.Y"
Bind ControlScreen   "c-f2"   := "CC$PlanetAt 2, UI.X, UI.Y"
Bind ControlScreen   "s-f3"   := "CC$PlanetAt 3, Chart.X, Chart.Y"
Bind ControlScreen   "c-f3"   := "CC$PlanetAt 3, UI.X, UI.Y"
Bind ControlScreen   "s-f8"   := "CC$PlanetAt 3, Chart.X, Chart.Y"
Bind ControlScreen   "c-f8"   := "CC$PlanetAt 3, UI.X, UI.Y"
Bind ControlScreen   "s-f4"   := "CC$GotoChart Chart.X, Chart.Y"
Bind ControlScreen   "c-f4"   := "CC$GotoChart UI.X, UI.Y"
Bind ControlScreen   "esc"    := "UI.GotoScreen 0"
Bind ControlScreen   "s-l"    := "CC$ListShips Chart.X, Chart.Y, 'a'"
Bind ControlScreen   "c-l"    := "CC$ListShips UI.X, UI.Y, 'a'"
Bind ControlScreen   "l"      := "CC$ListShips UI.X, UI.Y, 'a'"
Bind ControlScreen   "c-n"    := "CC$ListShipPrediction UI.X, UI.Y"
Bind ControlScreen   "f1"     := "CC$GotoScreen 1"
Bind ControlScreen   "f2"     := "CC$GotoScreen 2"
Bind ControlScreen   "f3"     := "CC$GotoScreen 3"
Bind ControlScreen   "f4"     := "CC$GotoChart Chart.X, Chart.Y"
Bind ControlScreen   "f5"     := "CC$PlanetInfo Chart.X, Chart.Y"
Bind ControlScreen   "s-f5"   := "CC$PlanetInfo Chart.X, Chart.Y"
Bind ControlScreen   "c-f5"   := "CC$PlanetInfo UI.X, UI.Y"
Bind ControlScreen   "f6"     := "CC$GotoScreen 6"
Bind ControlScreen   "s-f6"   := "CC$GotoScreen 6"
% Bind ControlScreen   "c-f6"   := "CC$GotoScreen 6"  -- how to do this?
Bind ControlScreen   "f7"     := "UI.Search"
Bind ControlScreen   "f10"    := "CC$GotoScreen 10"
Bind ControlScreen   "c-h"    := "UI.Search Chart.X & ', ' & Chart.Y, 'spbuo4'"
Bind ControlScreen   "alt-r"  := "CC$Reset Loc.X, Loc.Y"

Bind FleetScreen     "b"      := "CC$ChangeLeader"
Bind FleetScreen     "c"      := "CC$ShipCargo"
Bind FleetScreen     "d"      := "CC$TransferMulti 1"
Bind FleetScreen     "f"      := "CC$ChangeFCode"
Bind FleetScreen     "g"      := "CC$Give"
Bind FleetScreen     "j"      := "CC$FleetJoin"
Bind FleetScreen     "n"      := "CC$FleetName"
Bind FleetScreen     "p"      := "CC$FleetSplit"
Bind FleetScreen     "r"      := "Try CC$RemoteControl"
Bind FleetScreen     "s"      := "CC$ShipSpec"
Bind FleetScreen     "t"      := "CC$TowFleetMember"
Bind FleetScreen     "u"      := "CC$ShipUnload"
Bind FleetScreen     "x"      := "CC$ShipXchg"
Bind FleetScreen     "f1"     := "CC$GotoObject 1, Id"
Bind FleetScreen     "f6"     := "CC$GotoObject 6, Id"
Bind FleetScreen     "f9"     := "CC$SetShipComment"
Bind FleetScreen     "ins"    := "Try CC$AddToSim"
Bind FleetScreen     "c-ins"  := "CC$FleetAddToSim"
Bind FleetScreen     "del"    := "CC$FleetDeleteMember"
Bind FleetScreen     "c-del"  := "CC$FleetDeleteFleet"
Bind FleetScreen     "c-w"    := "CC$WithFleetWaypoint 'SetWaypoint UI.X, UI.Y', True"
Bind FleetScreen     "w"      := "CC$WithFleetWaypoint 'CC$ChangeSpeed', False"
Bind FleetScreen     "a"      := "CC$WithFleetWaypoint 'CC$ChangeWaypoint', False"
Bind FleetScreen     "dblclick" := "CC$WithFleetWaypoint 'SetWaypoint UI.X, UI.Y', True"
Bind FleetScreen     "h"      := "UI.Help 'pcc2:fleetscreen'"
Bind FleetScreen     "alt-h"  := "UI.Help 'pcc2:fleetscreen'"

Bind ShipLock        "."      := "CC$ToggleSelection"
Bind ShipLock        "f1"     := "CC$GotoObject 1, Id"
Bind ShipLock        "s-f1"   := "CC$GotoObject 1, Id"
Bind ShipLock        "c-f1"   := "CC$GotoObject 1, Id"
Bind ShipLock        "f6"     := "CC$GotoObject 6, Id"
Bind ShipLock        "s-f6"   := "CC$GotoObject 6, Id"
Bind ShipLock        "c-f6"   := "CC$GotoObject 6, Id"
Bind ShipLock        "esc"    := "CC$HidePanel"
Bind ShipLock        "c-n"    := "CC$ListShipPrediction UI.X, UI.Y, Id"
Bind PlanetLock      "."      := "CC$ToggleSelection"
Bind PlanetLock      "b"      := "CC$BaseLock"
Bind PlanetLock      "a"      := "Try CC$BuildStructures 0"
Bind PlanetLock      "esc"    := "CC$HidePanel"
Bind BaseLock        "p"      := "CC$PlanetLock"
Bind BaseLock        "esc"    := "CC$HidePanel"
Bind Starchart       "s-f9"   := "CC$SetPlanetCommentAt UI.X, UI.Y"
Bind Starchart       "c-f9"   := "CC$SetPlanetCommentAt UI.X, UI.Y"
Bind Starchart       "f7"     := "UI.Search"
Bind Starchart       "c-h"    := "UI.Search UI.X & ', ' & UI.Y, 'spbuo4'"
Bind Starchart       "esc"    := "UI.GotoScreen 0"
Bind Starchart       "f4"     := "UI.GotoScreen 0"
Bind Starchart       "f1"     := "CC$ListShips UI.X, UI.Y, 'e'"
Bind Starchart       "s-f1"   := "CC$ListShips UI.X, UI.Y, 'e'"
Bind Starchart       "c-f1"   := "CC$ListShips UI.X, UI.Y, 'e'"
Bind Starchart       "c-n"    := "CC$ListShipPrediction UI.X, UI.Y"
Bind Starchart       "f2"     := "CC$PlanetAt 2, UI.X, UI.Y"
Bind Starchart       "s-f2"   := "CC$PlanetAt 2, UI.X, UI.Y"
Bind Starchart       "c-f2"   := "CC$PlanetAt 2, UI.X, UI.Y"
Bind Starchart       "f3"     := "CC$PlanetAt 3, UI.X, UI.Y"
Bind Starchart       "s-f3"   := "CC$PlanetAt 3, UI.X, UI.Y"
Bind Starchart       "c-f3"   := "CC$PlanetAt 3, UI.X, UI.Y"
Bind Starchart       "f6"     := "CC$GotoScreen 6"
Bind Starchart       "s-f6"   := "CC$GotoScreen 6"
Bind Starchart       "c-f6"   := "CC$GotoScreen 6"
Bind Starchart       "f8"     := "CC$PlanetAt 3, UI.X, UI.Y"
Bind Starchart       "s-f8"   := "CC$PlanetAt 3, UI.X, UI.Y"
Bind Starchart       "c-f8"   := "CC$PlanetAt 3, UI.X, UI.Y"
Bind Starchart       "l"      := "CC$ListShips UI.X, UI.Y, 'a'"
Bind Starchart       "s-l"    := "CC$ListShips UI.X, UI.Y, 'a'"
Bind Starchart       "f5"     := "CC$PlanetInfo UI.X, UI.Y"
Bind Starchart       "s-f5"   := "CC$PlanetInfo UI.X, UI.Y"
Bind Starchart       "c-f5"   := "CC$PlanetInfo UI.X, UI.Y"

Bind Starchart       "left"   := "CC$ScannerMoveX -10"
Bind Starchart       "right"  := "CC$ScannerMoveX 10"
Bind Starchart       "up"     := "CC$ScannerMoveY 10"
Bind Starchart       "down"   := "CC$ScannerMoveY -10"
Bind Starchart       "s-left" := "CC$ScannerMoveX -1"
Bind Starchart       "s-right":= "CC$ScannerMoveX 1"
Bind Starchart       "s-up"   := "CC$ScannerMoveY 1"
Bind Starchart       "s-down" := "CC$ScannerMoveY -1"
Bind Starchart       "c-left" := "CC$ScannerMoveX -100"
Bind Starchart       "c-right":= "CC$ScannerMoveX 100"
Bind Starchart       "c-up"   := "CC$ScannerMoveY 100"
Bind Starchart       "c-down" := "CC$ScannerMoveY -100"
Bind Starchart       "h"      := "UI.Help 'pcc2:starchart'"
Bind Starchart       "alt-h"  := "UI.Help 'pcc2:starchart'"

Bind RaceScreen      "f1"     := "CC$GotoScreen 1"
Bind RaceScreen      "f2"     := "CC$GotoScreen 2"
Bind RaceScreen      "f3"     := "CC$GotoScreen 3"
Bind RaceScreen      "f4"     := "UI.GotoScreen 4"
Bind RaceScreen      "f6"     := "CC$GotoScreen 6"
Bind RaceScreen      "f7"     := "UI.Search"
Bind RaceScreen      "f10"    := "CC$GotoScreen 10"
Bind RaceScreen      "m"      := "CC$ViewInbox"
Bind RaceScreen      "w"      := "CC$SendMessage"
Bind RaceScreen      "c-s"    := "SaveGame"
Bind RaceScreen      "esc"    := "CC$ExitRace"
Bind RaceScreen      "h"      := "UI.Help 'pcc2:racescreen'"
Bind RaceScreen      "alt-h"  := "UI.Help 'pcc2:racescreen'"
Bind RaceScreen      "ctrl-o" := "CC$Settings"
Bind RaceScreen      "a"      := "UI.EditAlliances"
Bind RaceScreen      "alt-u"  := "CC$ViewUtilData"

Bind Global          "a-c"    := "UI.PopupConsole"
Bind Global          "a-k"    := "UI.KeymapInfo"
Bind Global          "a-left" := "CC$SwitchSelection -1"
Bind Global          "a-right" := "CC$SwitchSelection 1"
Bind Global          "a-."    := "UI.SelectionManager"
Bind Global          "quit"   := "CC$ExitClient"
Bind Global          "bs"     := "CC$PopScreenHistory"
Bind Global          "a-bs"   := "CC$ListScreenHistory"

% Document internally-handled keys.
% These keys are processed by PCC2 core.
% We define them here so they do not show up as unbound in keymap listings.
% Assigning them a non-atom number will achieve just that.
Bind AutoTaskScreen   "c-del" := 1, "c-n" := 1, "c-r" := 1, "c-s" := 1             % WAutoTaskScreen
Bind AutoTaskScreen   "del" := 1, "ins" := 1, "spc" := 1, "%" := 1, "k" := 1       % WAutoTaskScreen
Bind BaseScreen       "r" := 1                                                     % WBaseScreen
Bind BaseTaskScreen   "b" := 1, "1" := 1, "2" := 1, "3" := 1, "4" := 1             % WBaseAutoTaskCommandTile
Bind ControlScreen    "y" := 1, "alt-q"  := 1                                      % WControlScreen
Bind HistoryScreen    "tab" := 1, "s-tab" := 1, "c-tab" := 1, "s-c-tab" := 1       % WHistoryScreen
Bind PlanetTaskScreen "1" := 1, "2" := 1, "3" := 1                                 % WPlanetAutoTaskCommandTile
Bind RaceScreen       "alt-e" := 1, "alt-q" := 1, "alt-n" := 1                     % WPlayerScreen
Bind RaceScreen       "v" := 1, "b" := 1, "s" := 1, "t" := 1, "i" := 1, "g" := 1   % WPlayerScreen
Bind ShipScreen       "b" := 1, "f10" := 1                                         % WShipScreen
Bind ShipScreen       "tab" := 1, "s-tab" := 1, "c-tab" := 1, "s-c-tab" := 1       % WShipScreen
Bind ShipTaskScreen   "1" := 1, "2" := 1, "3" := 1, "4" := 1, "o" := 1             % WShipAutoTaskCommandTile
Bind ShipTaskScreen   "c-w" := 1                                                   % WShipAutoTaskScreen
Bind ShipTaskScreen   "g" := 1, "a" := 1, "e" := 1, "d" := 1                       % WShipAutoTaskCommandTile
Bind Starchart        "ret" := 1, "spc" := 1, "c-ret" := 1, "c-spc" := 1           % WStandardChartMode
Bind Starchart        "-" := 1, "wheeldn" := 1, "+" := 1, "wheelup" := 1           % WStandardChartMode
Bind Starchart        "tab" := 1, "s-tab" := 1, "c-tab" := 1, "s-c-tab" := 1       % WStandardChartMode
Bind Starchart        "pgup" := 1, "c-pgup" := 1, "pgdn" := 1, "c-pgdn" := 1       % WStandardChartMode
Bind Starchart        "s-pgup" := 1, "s-c-pgup" := 1, "s-pgdn" := 1, "s-c-pgdn" := 1
Bind Starchart        "f9" := 1, "alt-f9" := 1, "m" := 1, "c-m" := 1               % WStandardChartMode
Bind Starchart        "u" := 1, "i" := 1, "c" := 1, "t" := 1, "v" := 1, "p" := 1   % WStandardChartMode
Bind Starchart        "d" := 1, "r" := 1, "g" := 1, "del" := 1                     % WStandardChartMode
Bind Starchart        "alt-l" := 1, "alt-o" := 1, "alt-f" := 1, "x" := 1           % WStandardChartMode
Bind Starchart        "alt-m" := 1, "alt-a" := 1, "alt-d" := 1, "alt-i" := 1       % GChartOptions
Bind Starchart        "alt-v" := 1, "alt-t" := 1, "alt-s" := 1, "alt-b" := 1       % GChartOptions
Bind Starchart        "alt-u" := 1, "alt-p" := 1, "alt-w" := 1                     % GChartOptions

On EnterDirectory Do TryLoad "autoexec.q"

If System.GUI Then Print "[", System.Program, " ", System.Version, ", core.q ", CC$LibraryVersion, "]"
